==> amqp-client-1.8.0-sources.jar: com/rabbitmq/client/ConnectionFactory.java <==
//   The contents of this file are subject to the Mozilla Public License
//   Version 1.1 (the "License"); you may not use this file except in
//   compliance with the License. You may obtain a copy of the License at
//   http://www.mozilla.org/MPL/
//
//   Software distributed under the License is distributed on an "AS IS"
//   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//   License for the specific language governing rights and limitations
//   under the License.
//
//   The Original Code is RabbitMQ.
//
//   The Initial Developers of the Original Code are LShift Ltd,
//   Cohesive Financial Technologies LLC, and Rabbit Technologies Ltd.
//
//   Portions created before 22-Nov-2008 00:00:00 GMT by LShift Ltd,
//   Cohesive Financial Technologies LLC, or Rabbit Technologies Ltd
//   are Copyright (C) 2007-2008 LShift Ltd, Cohesive Financial
//   Technologies LLC, and Rabbit Technologies Ltd.
//
//   Portions created by LShift Ltd are Copyright (C) 2007-2010 LShift
//   Ltd. Portions created by Cohesive Financial Technologies LLC are
//   Copyright (C) 2007-2010 Cohesive Financial Technologies
//   LLC. Portions created by Rabbit Technologies Ltd are Copyright
//   (C) 2007-2010 Rabbit Technologies Ltd.
//
//   All Rights Reserved.
//
//   Contributor(s): ______________________________________.
//
package com.rabbitmq.client;

import java.io.IOException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;

import java.net.Socket;
import java.net.InetSocketAddress;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;

import com.rabbitmq.client.impl.AMQConnection;
import com.rabbitmq.client.impl.FrameHandler;
import com.rabbitmq.client.impl.SocketFrameHandler;

/**
 * Convenience "factory" class to facilitate opening a {@link Connection} to an AMQP broker.
 */

public class ConnectionFactory implements Cloneable {
    /** Default user name */
    public static final String DEFAULT_USER = "guest";

    /** Default password */
    public static final String DEFAULT_PASS = "guest";

    /** Default virtual host */
    public static final String DEFAULT_VHOST = "/";

    /** Default value for the desired maximum channel number; zero for
     * unlimited */
    public static final int DEFAULT_CHANNEL_MAX = 0;

    /** Default value for the desired maximum frame size; zero for
     * unlimited */
    public static final int DEFAULT_FRAME_MAX = 0;

    /** Default value for desired heartbeat interval; zero for none */
    public static final int DEFAULT_HEARTBEAT = 0;

    /** The default host to connect to */
    public static final String DEFAULT_HOST = "localhost";

    /** A constant that when passed as a port number causes the connection to use the default port */
    public static final int USE_DEFAULT_PORT = -1;

    /** The default port to use for AMQP connections when not using SSL */
    public static final int DEFAULT_AMQP_PORT = 5672;

    /** The default port to use for AMQP connections when using SSL */
    public static final int DEFAULT_AMQP_OVER_SSL_PORT = 5671;

    /**
     * The default SSL protocol (currently "SSLv3").
     */
    public static final String DEFAULT_SSL_PROTOCOL = "SSLv3";

    private String username                       = DEFAULT_USER;
    private String password                       = DEFAULT_PASS;
    private String virtualHost                    = DEFAULT_VHOST;
    private String host                           = DEFAULT_HOST;
    private int port                              = USE_DEFAULT_PORT;
    private int requestedChannelMax               = DEFAULT_CHANNEL_MAX;
    private int requestedFrameMax                 = DEFAULT_FRAME_MAX;
    private int requestedHeartbeat                = DEFAULT_HEARTBEAT;
    private Map<String, Object> _clientProperties = AMQConnection.defaultClientProperties();
    private SocketFactory factory                 = SocketFactory.getDefault();

    /**
     * Instantiate a ConnectionFactory with a default set of parameters.
     */
    public ConnectionFactory() {
    }

    /**
     *  @return the default host to use for connections
     */
    public String getHost() {
        return host;
    }

    /**
     *  @param host the default host to use for connections
     */
    public void setHost(String host) {
        this.host = host;
    }

    private int portOrDefault(int port){
        if(port != USE_DEFAULT_PORT) return port;
        else if(isSSL()) return DEFAULT_AMQP_OVER_SSL_PORT;
        else return DEFAULT_AMQP_PORT;
    }

    /**
     *  @return the default port to use for connections
     */
    public int getPort() {
        return portOrDefault(port);
    }

    /**
     *  @return the default port to use for connections
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * Retrieve the user name.
     * @return the AMQP user name to use when connecting to the broker
     */
    public String getUsername() {
        return this.username;
    }

    /**
     * Set the user name.
     * @param username the AMQP user name to use when connecting to the broker
     */
    public void setUsername(String username) {
        this.username = username;
    }

    /**
     * Retrieve the password.
     * @return the password to use when connecting to the broker
     */
    public String getPassword() {
        return this.password;
    }

    /**
     * Set the password.
     * @param password the password to use when connecting to the broker
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * Retrieve the virtual host.
     * @return the virtual host to use when connecting to the broker
     */
    public String getVirtualHost() {
        return this.virtualHost;
    }

    /**
     * Set the virtual host.
     * @param virtualHost the virtual host to use when connecting to the broker
     */
    public void setVirtualHost(String virtualHost) {
        this.virtualHost = virtualHost;
    }

    /**
     * Retrieve the requested maximum channel number
     * @return the initially requested maximum channel number; zero for unlimited
     */
    public int getRequestedChannelMax() {
        return this.requestedChannelMax;
    }

    /**
     * Set the requested maximum channel number
     * @param requestedChannelMax initially requested maximum channel number; zero for unlimited
     */
    public void setRequestedChannelMax(int requestedChannelMax) {
        this.requestedChannelMax = requestedChannelMax;
    }

    /**
     * Retrieve the requested maximum frame size
     * @return the initially requested maximum frame size, in octets; zero for unlimited
     */
    public int getRequestedFrameMax() {
        return this.requestedFrameMax;
    }

    /**
     * Set the requested maximum frame size
     * @param requestedFrameMax initially requested maximum frame size, in octets; zero for unlimited
     */
    public void setRequestedFrameMax(int requestedFrameMax) {
        this.requestedFrameMax = requestedFrameMax;
    }

    /**
     * Retrieve the requested heartbeat interval.
     * @return the initially requested heartbeat interval, in seconds; zero for none
     */
    public int getRequestedHeartbeat() {
        return this.requestedHeartbeat;
    }

    /**
     * Set the requested heartbeat.
     * @param requestedHeartbeat the initially requested heartbeat interval, in seconds; zero for none
     */
    public void setRequestedHeartbeat(int requestedHeartbeat) {
        this.requestedHeartbeat = requestedHeartbeat;
    }

    /**
     * Retrieve the currently-configured table of client properties
     * that will be sent to the server during connection
     * startup. Clients may add, delete, and alter keys in this
     * table. Such changes will take effect when the next new
     * connection is started using this factory.
     * @return the map of client properties
     * @see setClientProperties()
     */
    public Map<String, Object> getClientProperties() {
        return _clientProperties;
    }

    /**
     * Replace the table of client properties that will be sent to the
     * server during subsequent connection startups.
     * @param clientProperties the map of extra client properties
     * @see getClientProperties()
     */
    public void setClientProperties(Map<String, Object> clientProperties) {
        _clientProperties = clientProperties;
    }

    /**
     * Retrieve the socket factory used to make connections with.
     */
    public SocketFactory getSocketFactory() {
        return this.factory;
    }

    /**
     * Set the socket factory used to make connections with. Can be
     * used to enable SSL connections by passing in a
     * javax.net.ssl.SSLSocketFactory instance.
     *
     * @see #useSslProtocol
     */
    public void setSocketFactory(SocketFactory factory) {
        this.factory = factory;
    }

    public boolean isSSL(){
        return getSocketFactory() instanceof SSLSocketFactory;
    }

    /**
     * Convenience method for setting up a SSL socket factory, using
     * the DEFAULT_SSL_PROTOCOL and a trusting TrustManager.
     */
    public void useSslProtocol()
        throws NoSuchAlgorithmException, KeyManagementException
    {
        useSslProtocol(DEFAULT_SSL_PROTOCOL);
    }

    /**
     * Convenience method for setting up a SSL socket factory, using
     * the DEFAULT_SSL_PROTOCOL and a trusting TrustManager.
     */
    public void useSslProtocol(String protocol)
        throws NoSuchAlgorithmException, KeyManagementException
    {
        useSslProtocol(protocol, new NullTrustManager());
    }

    /**
     * Convenience method for setting up an SSL socket factory.
     * Pass in the SSL protocol to use, e.g. "TLS" or "SSLv3".
     *
     * @param protocol SSL protocol to use.
     */
    public void useSslProtocol(String protocol, TrustManager trustManager)
        throws NoSuchAlgorithmException, KeyManagementException
    {
        SSLContext c = SSLContext.getInstance(protocol);
        c.init(null, new TrustManager[] { trustManager }, null);
        useSslProtocol(c);
    }

    /**
     * Convenience method for setting up an SSL socket factory.
     * Pass in an initialized SSLContext.
     *
     * @param context An initialized SSLContext
     */
    public void useSslProtocol(SSLContext context)
    {
        setSocketFactory(context.getSocketFactory());
    }

    protected FrameHandler createFrameHandler(Address addr)
        throws IOException {

        String hostName = addr.getHost();
        int portNumber = portOrDefault(addr.getPort());
        Socket socket = factory.createSocket();
        configureSocket(socket);
        socket.connect(new InetSocketAddress(hostName, portNumber));
        return new SocketFrameHandler(socket);
    }

    /**
     *  Provides a hook to insert custom configuration of the sockets used
     *  to connect to an AMQP server before they connect. 
     *
     *  The default behaviour of this method is to disable Nagle's algorithm to get
     *  more consistently low latency.
     *  However it may be overridden freely and there is no requirement to retain
     *  this behaviour. 
     *
     *  @param socket The socket that is to be used for the Connection
     */
    protected void configureSocket(Socket socket) throws IOException{
        // disable Nagle's algorithm, for more consistently low latency
        socket.setTcpNoDelay(true);
    }

    private Connection newConnection(Address[] addresses,
                                     int maxRedirects,
                                     Map<Address,Integer> redirectAttempts)
        throws IOException
    {
        IOException lastException = null;
        for (Address address : addresses) {
            Address[] lastKnownAddresses = new Address[0];
            try {
                while(true) {
                    FrameHandler frameHandler = createFrameHandler(address);
                    Integer redirectCount = redirectAttempts.get(address);
                    if (redirectCount == null)
                        redirectCount = 0;
                    boolean allowRedirects = redirectCount < maxRedirects;
                    try {
                        AMQConnection conn = new AMQConnection(this,
                                    frameHandler);
                        conn.start(!allowRedirects);
                        return conn;
                    } catch (RedirectException e) {
                        if (!allowRedirects) {
                            //this should never happen with a well-behaved server
                            throw new IOException("server ignored 'insist'");
                        } else {
                            redirectAttempts.put(address, redirectCount+1);
                            lastKnownAddresses = e.getKnownAddresses();
                            address = e.getAddress();
                            //TODO: we may want to log redirection attempts.
                        }
                    }
                }
            } catch (IOException e) {
                lastException = e;
                if (lastKnownAddresses.length > 0) {
                    // If there aren't any, don't bother trying, since
                    // a recursive call with empty lastKnownAddresses
                    // will cause our lastException to be stomped on
                    // by an uninformative IOException. See bug 16273.
                    try {
                        return newConnection(lastKnownAddresses,
                                             maxRedirects,
                                             redirectAttempts);
                    } catch (IOException e1) {
                        lastException = e1;
                    }
                }
            }
        }

        if (lastException == null) {
            throw new IOException("failed to connect");
        } else {
            throw lastException;
        }
    }

    /**
     * Create a new broker connection
     * @return an interface to the connection
     * @throws IOException if it encounters a problem
     */
    public Connection newConnection() throws IOException {
        return newConnection(new Address[] {
                                 new Address(getHost(), getPort())
                             },
                             0,
                             new HashMap<Address,Integer>());
    }


    @Override public ConnectionFactory clone(){
        try {
            return (ConnectionFactory)super.clone(); 
        } catch (CloneNotSupportedException e) {
            throw new Error(e);
        }
    }
}

==> amqp-client-1.8.1-sources.jar: com/rabbitmq/client/ConnectionFactory.java <==
//   The contents of this file are subject to the Mozilla Public License
//   Version 1.1 (the "License"); you may not use this file except in
//   compliance with the License. You may obtain a copy of the License at
//   http://www.mozilla.org/MPL/
//
//   Software distributed under the License is distributed on an "AS IS"
//   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//   License for the specific language governing rights and limitations
//   under the License.
//
//   The Original Code is RabbitMQ.
//
//   The Initial Developers of the Original Code are LShift Ltd,
//   Cohesive Financial Technologies LLC, and Rabbit Technologies Ltd.
//
//   Portions created before 22-Nov-2008 00:00:00 GMT by LShift Ltd,
//   Cohesive Financial Technologies LLC, or Rabbit Technologies Ltd
//   are Copyright (C) 2007-2008 LShift Ltd, Cohesive Financial
//   Technologies LLC, and Rabbit Technologies Ltd.
//
//   Portions created by LShift Ltd are Copyright (C) 2007-2010 LShift
//   Ltd. Portions created by Cohesive Financial Technologies LLC are
//   Copyright (C) 2007-2010 Cohesive Financial Technologies
//   LLC. Portions created by Rabbit Technologies Ltd are Copyright
//   (C) 2007-2010 Rabbit Technologies Ltd.
//
//   All Rights Reserved.
//
//   Contributor(s): ______________________________________.
//
package com.rabbitmq.client;

import java.io.IOException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;

import java.net.Socket;
import java.net.InetSocketAddress;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;

import com.rabbitmq.client.impl.AMQConnection;
import com.rabbitmq.client.impl.FrameHandler;
import com.rabbitmq.client.impl.SocketFrameHandler;

/**
 * Convenience "factory" class to facilitate opening a {@link Connection} to an AMQP broker.
 */

public class ConnectionFactory implements Cloneable {
    /** Default user name */
    public static final String DEFAULT_USER = "guest";

    /** Default password */
    public static final String DEFAULT_PASS = "guest";

    /** Default virtual host */
    public static final String DEFAULT_VHOST = "/";

    /** Default value for the desired maximum channel number; zero for
     * unlimited */
    public static final int DEFAULT_CHANNEL_MAX = 0;

    /** Default value for the desired maximum frame size; zero for
     * unlimited */
    public static final int DEFAULT_FRAME_MAX = 0;

    /** Default value for desired heartbeat interval; zero for none */
    public static final int DEFAULT_HEARTBEAT = 0;

    /** The default host to connect to */
    public static final String DEFAULT_HOST = "localhost";

    /** A constant that when passed as a port number causes the connection to use the default port */
    public static final int USE_DEFAULT_PORT = -1;

    /** The default port to use for AMQP connections when not using SSL */
    public static final int DEFAULT_AMQP_PORT = 5672;

    /** The default port to use for AMQP connections when using SSL */
    public static final int DEFAULT_AMQP_OVER_SSL_PORT = 5671;

    /**
     * The default SSL protocol (currently "SSLv3").
     */
    public static final String DEFAULT_SSL_PROTOCOL = "SSLv3";

    private String username                       = DEFAULT_USER;
    private String password                       = DEFAULT_PASS;
    private String virtualHost                    = DEFAULT_VHOST;
    private String host                           = DEFAULT_HOST;
    private int port                              = USE_DEFAULT_PORT;
    private int requestedChannelMax               = DEFAULT_CHANNEL_MAX;
    private int requestedFrameMax                 = DEFAULT_FRAME_MAX;
    private int requestedHeartbeat                = DEFAULT_HEARTBEAT;
    private Map<String, Object> _clientProperties = AMQConnection.defaultClientProperties();
    private SocketFactory factory                 = SocketFactory.getDefault();

    /**
     * Instantiate a ConnectionFactory with a default set of parameters.
     */
    public ConnectionFactory() {
    }

    /**
     *  @return the default host to use for connections
     */
    public String getHost() {
        return host;
    }

    /**
     *  @param host the default host to use for connections
     */
    public void setHost(String host) {
        this.host = host;
    }

    private int portOrDefault(int port){
        if(port != USE_DEFAULT_PORT) return port;
        else if(isSSL()) return DEFAULT_AMQP_OVER_SSL_PORT;
        else return DEFAULT_AMQP_PORT;
    }

    /**
     *  @return the default port to use for connections
     */
    public int getPort() {
        return portOrDefault(port);
    }

    /**
     * Set the target port.
     * @param port the default port to use for connections
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * Retrieve the user name.
     * @return the AMQP user name to use when connecting to the broker
     */
    public String getUsername() {
        return this.username;
    }

    /**
     * Set the user name.
     * @param username the AMQP user name to use when connecting to the broker
     */
    public void setUsername(String username) {
        this.username = username;
    }

    /**
     * Retrieve the password.
     * @return the password to use when connecting to the broker
     */
    public String getPassword() {
        return this.password;
    }

    /**
     * Set the password.
     * @param password the password to use when connecting to the broker
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * Retrieve the virtual host.
     * @return the virtual host to use when connecting to the broker
     */
    public String getVirtualHost() {
        return this.virtualHost;
    }

    /**
     * Set the virtual host.
     * @param virtualHost the virtual host to use when connecting to the broker
     */
    public void setVirtualHost(String virtualHost) {
        this.virtualHost = virtualHost;
    }

    /**
     * Retrieve the requested maximum channel number
     * @return the initially requested maximum channel number; zero for unlimited
     */
    public int getRequestedChannelMax() {
        return this.requestedChannelMax;
    }

    /**
     * Set the requested maximum channel number
     * @param requestedChannelMax initially requested maximum channel number; zero for unlimited
     */
    public void setRequestedChannelMax(int requestedChannelMax) {
        this.requestedChannelMax = requestedChannelMax;
    }

    /**
     * Retrieve the requested maximum frame size
     * @return the initially requested maximum frame size, in octets; zero for unlimited
     */
    public int getRequestedFrameMax() {
        return this.requestedFrameMax;
    }

    /**
     * Set the requested maximum frame size
     * @param requestedFrameMax initially requested maximum frame size, in octets; zero for unlimited
     */
    public void setRequestedFrameMax(int requestedFrameMax) {
        this.requestedFrameMax = requestedFrameMax;
    }

    /**
     * Retrieve the requested heartbeat interval.
     * @return the initially requested heartbeat interval, in seconds; zero for none
     */
    public int getRequestedHeartbeat() {
        return this.requestedHeartbeat;
    }

    /**
     * Set the requested heartbeat.
     * @param requestedHeartbeat the initially requested heartbeat interval, in seconds; zero for none
     */
    public void setRequestedHeartbeat(int requestedHeartbeat) {
        this.requestedHeartbeat = requestedHeartbeat;
    }

    /**
     * Retrieve the currently-configured table of client properties
     * that will be sent to the server during connection
     * startup. Clients may add, delete, and alter keys in this
     * table. Such changes will take effect when the next new
     * connection is started using this factory.
     * @return the map of client properties
     * @see #setClientProperties
     */
    public Map<String, Object> getClientProperties() {
        return _clientProperties;
    }

    /**
     * Replace the table of client properties that will be sent to the
     * server during subsequent connection startups.
     * @param clientProperties the map of extra client properties
     * @see #getClientProperties
     */
    public void setClientProperties(Map<String, Object> clientProperties) {
        _clientProperties = clientProperties;
    }

    /**
     * Retrieve the socket factory used to make connections with.
     */
    public SocketFactory getSocketFactory() {
        return this.factory;
    }

    /**
     * Set the socket factory used to make connections with. Can be
     * used to enable SSL connections by passing in a
     * javax.net.ssl.SSLSocketFactory instance.
     *
     * @see #useSslProtocol
     */
    public void setSocketFactory(SocketFactory factory) {
        this.factory = factory;
    }

    public boolean isSSL(){
        return getSocketFactory() instanceof SSLSocketFactory;
    }

    /**
     * Convenience method for setting up a SSL socket factory, using
     * the DEFAULT_SSL_PROTOCOL and a trusting TrustManager.
     */
    public void useSslProtocol()
        throws NoSuchAlgorithmException, KeyManagementException
    {
        useSslProtocol(DEFAULT_SSL_PROTOCOL);
    }

    /**
     * Convenience method for setting up a SSL socket factory, using
     * the DEFAULT_SSL_PROTOCOL and a trusting TrustManager.
     */
    public void useSslProtocol(String protocol)
        throws NoSuchAlgorithmException, KeyManagementException
    {
        useSslProtocol(protocol, new NullTrustManager());
    }

    /**
     * Convenience method for setting up an SSL socket factory.
     * Pass in the SSL protocol to use, e.g. "TLS" or "SSLv3".
     *
     * @param protocol SSL protocol to use.
     */
    public void useSslProtocol(String protocol, TrustManager trustManager)
        throws NoSuchAlgorithmException, KeyManagementException
    {
        SSLContext c = SSLContext.getInstance(protocol);
        c.init(null, new TrustManager[] { trustManager }, null);
        useSslProtocol(c);
    }

    /**
     * Convenience method for setting up an SSL socket factory.
     * Pass in an initialized SSLContext.
     *
     * @param context An initialized SSLContext
     */
    public void useSslProtocol(SSLContext context)
    {
        setSocketFactory(context.getSocketFactory());
    }

    protected FrameHandler createFrameHandler(Address addr)
        throws IOException {

        String hostName = addr.getHost();
        int portNumber = portOrDefault(addr.getPort());
        Socket socket = factory.createSocket();
        configureSocket(socket);
        socket.connect(new InetSocketAddress(hostName, portNumber));
        return createFrameHandler(socket);
    }

    protected FrameHandler createFrameHandler(Socket sock)
        throws IOException
    {
        return new SocketFrameHandler(sock);
    }

    /**
     *  Provides a hook to insert custom configuration of the sockets used
     *  to connect to an AMQP server before they connect. 
     *
     *  The default behaviour of this method is to disable Nagle's algorithm to get
     *  more consistently low latency.
     *  However it may be overridden freely and there is no requirement to retain
     *  this behaviour. 
     *
     *  @param socket The socket that is to be used for the Connection
     */
    protected void configureSocket(Socket socket) throws IOException{
        // disable Nagle's algorithm, for more consistently low latency
        socket.setTcpNoDelay(true);
    }

    private Connection newConnection(Address[] addresses,
                                     int maxRedirects,
                                     Map<Address,Integer> redirectAttempts)
        throws IOException
    {
        IOException lastException = null;
        for (Address address : addresses) {
            Address[] lastKnownAddresses = new Address[0];
            try {
                while(true) {
                    FrameHandler frameHandler = createFrameHandler(address);
                    Integer redirectCount = redirectAttempts.get(address);
                    if (redirectCount == null)
                        redirectCount = 0;
                    boolean allowRedirects = redirectCount < maxRedirects;
                    try {
                        AMQConnection conn = new AMQConnection(this,
                                    frameHandler);
                        conn.start(!allowRedirects);
                        return conn;
                    } catch (RedirectException e) {
                        if (!allowRedirects) {
                            //this should never happen with a well-behaved server
                            throw new IOException("server ignored 'insist'");
                        } else {
                            redirectAttempts.put(address, redirectCount+1);
                            lastKnownAddresses = e.getKnownAddresses();
                            address = e.getAddress();
                            //TODO: we may want to log redirection attempts.
                        }
                    }
                }
            } catch (IOException e) {
                lastException = e;
                if (lastKnownAddresses.length > 0) {
                    // If there aren't any, don't bother trying, since
                    // a recursive call with empty lastKnownAddresses
                    // will cause our lastException to be stomped on
                    // by an uninformative IOException. See bug 16273.
                    try {
                        return newConnection(lastKnownAddresses,
                                             maxRedirects,
                                             redirectAttempts);
                    } catch (IOException e1) {
                        lastException = e1;
                    }
                }
            }
        }

        if (lastException == null) {
            throw new IOException("failed to connect");
        } else {
            throw lastException;
        }
    }

    /**
     * Create a new broker connection
     * @return an interface to the connection
     * @throws IOException if it encounters a problem
     */
    public Connection newConnection() throws IOException {
        return newConnection(new Address[] {
                                 new Address(getHost(), getPort())
                             },
                             0,
                             new HashMap<Address,Integer>());
    }


    @Override public ConnectionFactory clone(){
        try {
            return (ConnectionFactory)super.clone(); 
        } catch (CloneNotSupportedException e) {
            throw new Error(e);
        }
    }
}
